"""
Tests for ``flickr_api.downloader``.
"""

import hashlib
from pathlib import Path
import time

import httpx
import pytest

from flickr_api import download_file


def test_download_photo(vcr_cassette: str, tmp_path: Path) -> None:
    """
    Download a photo from Flickr and check it's downloaded correctly.
    """
    result = download_file(
        url="https://live.staticflickr.com/65535/53574198477_fba34d20ca_c_d.jpg",
        download_dir=tmp_path,
        base_name="53574198477",
    )

    assert result == {
        "path": tmp_path / "53574198477.jpg",
        "content_type": "image/jpeg",
    }

    assert result["path"].exists()
    assert result["path"].stat().st_size == 126058
    assert (
        hashlib.md5(result["path"].read_bytes()).hexdigest()
        == "392b2e74d29ff90bb707658d422d14ad"
    )


def test_download_video(vcr_cassette: str, tmp_path: Path) -> None:
    """
    Download a video from Flickr and check it's downloaded correctly.
    """
    result = download_file(
        url="https://www.flickr.com/photos/straytoaster/51572201979/play/360p/6737dcd2a7/",
        download_dir=tmp_path,
        base_name="51572201979",
    )

    assert result == {"path": tmp_path / "51572201979.mp4", "content_type": "video/mp4"}


def test_not_found_is_error(vcr_cassette: str, tmp_path: Path) -> None:
    """
    Trying to fetch a Flickr URL that doesn't exist throws an immediate
    404 error.
    """
    t0 = time.time()

    with pytest.raises(httpx.HTTPStatusError):
        download_file(
            url="https://live.staticflickr.com/65535/doesnotexist.jpg",
            download_dir=tmp_path,
            base_name="doesnotexist",
        )

    # Check that less than 5 seconds elapsed -- we weren't waiting for
    # the library to retry anything.
    assert time.time() - t0 < 5


def test_an_unrecognised_content_type_is_downloaded_sans_extension(
    vcr_cassette: str, tmp_path: Path
) -> None:
    """
    If the URL has an unrecognised Content-Type, it's downloaded
    without an extension.
    """
    result = download_file(
        url="https://flickr.com",
        download_dir=tmp_path,
        base_name="homepage",
    )

    assert result == {
        "path": tmp_path / "homepage",
        "content_type": "text/html",
    }



from .api import FlickrApi
from .downloader import download_file
from .exceptions import (
    FlickrApiException,
    InsufficientPermissionsToComment,
    InvalidApiKey,
    InvalidXmlException,
    PermissionDenied,
    ResourceNotFound,
    LicenseNotFound,
    UnrecognisedFlickrApiException,
    UserDeleted,
)


__version__ = "3.11.3"


__all__ = [
    "download_file",
    "FlickrApi",
    "FlickrApiException",
    "ResourceNotFound",
    "InvalidApiKey",
    "InvalidXmlException",
    "LicenseNotFound",
    "InsufficientPermissionsToComment",
    "PermissionDenied",
    "PhotoContext",
    "UnrecognisedFlickrApiException",
    "UserDeleted",
    "__version__",
]



"""
Fixtures for creating instances of the FlickrApi class in tests.

To use these fixtures, add the following lines to the `conftest.py` file
in your project:

    from flickr_api.fixtures import flickr_api, flickr_oauth_api

    __all__ = ["flickr_api", "flickr_oauth_api"]

"""

from collections.abc import Iterator
import os
import typing

import httpx
from nitrate.cassettes import cassette_name
import pytest
import vcr

from flickr_api import FlickrApi


__all__ = ["cassette_name", "flickr_api", "flickr_oauth_api"]


def check_for_invalid_api_key(response: typing.Any) -> typing.Any:
    """
    Before we record a new response to a cassette, check if it's
    a Flickr API response telling us we're missing an API key -- that
    means we didn't set up the test correctly.

    If so, give the developer an instruction explaining what to do next.
    """
    try:
        body: bytes = response["body"]["string"]
    except KeyError:
        body = response["content"]

    is_error_response = body == (
        b'<?xml version="1.0" encoding="utf-8" ?>\n'
        b'<rsp stat="fail">\n\t'
        b'<err code="100" msg="Invalid API Key (Key has invalid format)" />\n'
        b"</rsp>\n"
    )
    skip_invalid_api_key_check = os.environ.get("SKIP_INVALID_API_KEY_CHECK") == "true"

    if is_error_response and not skip_invalid_api_key_check:
        raise RuntimeError(
            "You tried to record a new call to the Flickr API, \n"
            "but the tests didn't get a valid API key.\n"
            "\n"
            "Either:\n"
            "1. Pass an API key as an env var FLICKR_API_KEY=ae84…,\n"
            "   and re-run the test, or\n"
            "2. Pass an env var SKIP_INVALID_API_KEY_CHECK=true, if you \n"
            "   want to record an API call with an invalid key."
        )

    return response


@pytest.fixture
def flickr_api(cassette_name: str) -> Iterator[FlickrApi]:
    """
    Create an instance of the FlickrApi class for use in tests.

    This API is authenticated using an API key, which means it acts as
    a public user of the site -- it has no access to sensitive information.

    This API will record its interactions as "cassettes" using vcr.py --
    that is, the request and response are saved to YAML files which can
    be checked into the repo, then replayed offline (e.g. in CI tests).
    The cassette is redacted to ensure it does not contain our API key.
    """
    with vcr.use_cassette(
        cassette_name,
        cassette_library_dir="tests/fixtures/cassettes",
        filter_query_parameters=["api_key"],
        decode_compressed_response=True,
        before_record_response=check_for_invalid_api_key,
    ):
        client = httpx.Client(
            params={"api_key": os.environ.get("FLICKR_API_KEY", "<REDACTED>")},
            headers={
                "User-Agent": "flickr-photos-api <hello@flickr.org>",
                #
                # This forces the connection to close immediately after the
                # API returns a response, which avoids any warnings about
                # unclosed sockets
                "Connection": "Close",
            },
        )

        yield FlickrApi(client)


def check_for_oauth_token(response: typing.Any) -> typing.Any:
    """
    Before we record a new response to a cassette, check if it's
    a Flickr API response telling us we're missing an API key -- that
    means we didn't set up the test correctly.

    If so, give the developer an instruction explaining what to do next.
    """
    try:
        body: bytes = response["body"]["string"]
    except KeyError:  # pragma: no cover
        body = response["content"]

    is_error_response = body == (
        b'<?xml version="1.0" encoding="utf-8" ?>\n'
        b'<rsp stat="fail">\n\t<err code="98" msg="Invalid auth token" />\n</rsp>'
        b"\n"
    )
    skip_invalid_api_key_check = os.environ.get("SKIP_INVALID_API_KEY_CHECK") == "true"

    if is_error_response and not skip_invalid_api_key_check:
        raise RuntimeError(
            "You tried to record a new call to the Flickr API, \n"
            "but the tests didn't get a valid OAuth credentials.\n"
            "\n"
            "Either:\n"
            "1. Pass OAuth credentials in the following env vars:\n"
            "      - FLICKR_CLIENT_KEY\n"
            "      - FLICKR_CLIENT_SECRET\n"
            "      - FLICKR_OAUTH_TOKEN\n"
            "      - FLICKR_OAUTH_TOKEN_SECRET"
            "   and re-run the test, or\n"
            "2. Pass an env var SKIP_INVALID_API_KEY_CHECK=true, if you \n"
            "   want to record an API call with invalid credentials."
        )

    return response


@pytest.fixture
def flickr_oauth_api(cassette_name: str) -> Iterator[FlickrApi]:
    """
    Create an instance of the FlickrApi class for use in tests.

    This API is authenticated using an OAuth token, which means it acts as
    if it's logged in as this OAuth user, and has access to potentially
    private information.

    This API will record its interactions as "cassettes" using vcr.py --
    that is, the request and response are saved to YAML files which can
    be checked into the repo, then replayed offline (e.g. in CI tests).
    The cassette is redacted to ensure it does not contain OAuth credentials.
    """
    from authlib.integrations.httpx_client import OAuth1Client

    client_id = os.environ.get("FLICKR_CLIENT_KEY", "CLIENT_KEY")
    client_secret = os.environ.get("FLICKR_CLIENT_SECRET", "CLIENT_SECRET")
    token = os.environ.get("FLICKR_OAUTH_TOKEN", "OAUTH_TOKEN")
    token_secret = os.environ.get("FLICKR_OAUTH_TOKEN_SECRET", "OAUTH_TOKEN_SECRET")

    with vcr.use_cassette(
        cassette_name,
        cassette_library_dir="tests/fixtures/cassettes",
        filter_query_parameters=[
            "oauth_consumer_key",
            "oauth_nonce",
            "oauth_signature",
            "oauth_signature_method",
            "oauth_timestamp",
            "oauth_token",
            "oauth_verifier",
            "oauth_version",
        ],
        decode_compressed_response=True,
        before_record_response=check_for_oauth_token,
    ):
        client = OAuth1Client(
            client_id=client_id,
            client_secret=client_secret,
            signature_type="QUERY",
            token=token,
            token_secret=token_secret,
            headers={
                "User-Agent": "flickr-photos-api <hello@flickr.org>",
                #
                # This forces the connection to close immediately after the
                # API returns a response, which avoids any warnings about
                # unclosed sockets
                "Connection": "Close",
            },
        )

        yield FlickrApi(client)



"""
Models related to Flickr user accounts.
"""

from datetime import datetime
import typing


__all__ = ["ProfileInfo", "User", "UserInfo"]


class User(typing.TypedDict):
    """
    Basic information about a user.  This is enough to identify the
    owner of a photo and display an attribution link.
    """

    id: str
    username: str
    realname: str | None
    path_alias: str | None
    photos_url: str
    profile_url: str
    is_deleted: typing.NotRequired[bool]


class UserInfo(User):
    """
    A user with extra information, as returned from ``user_info``.

    This includes:

    * their profile description (if any)
    * their location (if any)
    * the number of photos they've uploaded
    * their buddy icon URL
    * whether they have Flickr Pro, and if so, when it expires

    The ``pro_account_expires`` field will be present if and only if
    the user has a Pro account.
    """

    description: str | None
    location: str | None
    count_photos: int
    buddy_icon_url: str

    # Note: I tried expressing the relationship ``pro_account_expires``
    # requires ``has_pro_account = True`` in the type system, but
    # it required a Union type that broke subclassing, i.e.
    #
    #     class UserInfoWithPro(…)
    #         has_pro_account: typing.Literal[True]
    #         pro_account_expires: datetime
    #
    #     class UserInfoWithoutPro(…)
    #         has_pro_account: typing.Literal[False]
    #
    #     UserInfo = UserInfoWithPro | UserInfoWithoutPro
    #
    # but this prevents subclassing this type, which we do in the
    # Commons Explorer -- so it's a convention rather than enforced
    # by the type system.
    has_pro_account: bool
    pro_account_expires: typing.NotRequired[datetime]


class ProfileInfo(typing.TypedDict):
    """
    A user's profile information.
    """

    id: str
    join_date: datetime

    occupation: str | None
    hometown: str | None

    showcase_album_id: str
    showcase_album_title: str

    first_name: str | None
    last_name: str | None
    email: str | None
    profile_description: str | None
    city: str | None
    country: str | None

    facebook: str | None
    twitter: str | None
    tumblr: str | None
    instagram: str | None
    pinterest: str | None



"""
Tests for `flickr_api.parsers`.
"""

import pytest

from flickr_api.parsers import parse_date_taken, parse_machine_tags, parse_safety_level


def test_unrecognised_date_granularity_is_error() -> None:
    """
    If you pass an unrecognised date taken granularity,
    it throws ``ValueError``.
    """
    with pytest.raises(ValueError, match="Unrecognised date granularity"):
        parse_date_taken(value="2017-02-17 00:00:00", granularity="-1", unknown=False)


def test_unrecognised_safety_level_is_error() -> None:
    """
    Parsing an unrecognised value as a safety level throws ``ValueError``.
    """
    with pytest.raises(ValueError, match="Unrecognised safety level"):
        parse_safety_level("-1")


class TestParseMachine_Tags:
    """
    Tests for `parse_machine_tags`.
    """

    def test_empty_tags_is_empty_machine_tags(self) -> None:
        """
        If there are no tags, there are no machine tags.
        """
        assert parse_machine_tags(tags=[]) == {}

    def test_keyword_tags_are_not_machine_tags(self) -> None:
        """
        If all the tags are "keywords" (unstructured text) then there
        are no machine tags.
        """
        tags = ["Natural history", "Periodicals", "Physical sciences"]
        assert parse_machine_tags(tags) == {}

    def test_get_single_machine_tag(self) -> None:
        """
        If some tags are machine tags and some aren't, only the machine tags
        are returned.
        """
        tags = [
            "Natural history",
            "Periodicals",
            "Physical sciences",
            "bhl:page=33665645",
        ]
        assert parse_machine_tags(tags) == {"bhl:page": ["33665645"]}



"""
When we call the Flickr API, sometimes the HTTP call fails with a retryable error.
For example:

*   An HTTP 500 Internal Server Error response
*   An HTTP 429 Too Many Requests response
*   A connection timeout or failure

In these cases, we can safely retry the request after a short delay and we
should get a successful response.

However, we don't want to retry all errors. For example, an HTTP 404 Not Found
is a permanent error that we should fail immediately, and not retry.

This file contains a function for deciding if a particular error can be retried,
or if we should give up immediately.
"""

import httpx

from .exceptions import InvalidXmlException, UnrecognisedFlickrApiException


__all__ = ["is_retryable"]


def is_retryable(exc: BaseException) -> bool:
    """
    Returns True if this is an exception we can safely retry (i.e. flaky
    or transient errors that might return a different result), or
    False otherwise.
    """
    if isinstance(exc, httpx.HTTPStatusError):
        if exc.response.status_code == 429 or exc.response.status_code >= 500:
            return True

    if isinstance(
        exc,
        (
            httpx.ConnectError,
            httpx.ConnectTimeout,
            httpx.ReadError,
            httpx.ReadTimeout,
            httpx.RemoteProtocolError,
            InvalidXmlException,
        ),
    ):
        return True

    # Sometimes we get an error from the Flickr API like:
    #
    #     <err
    #       code="201"
    #       msg="Sorry, the Flickr API service is not currently available."
    #     />
    #
    # but this indicates a flaky connection rather than a genuine failure.
    #
    # We've seen similar with code "0", so we match on the error message
    # rather than the code.
    if (
        isinstance(exc, UnrecognisedFlickrApiException)
        and isinstance(exc.args[0], dict)
        and exc.args[0].get("msg")
        == "Sorry, the Flickr API service is not currently available."
    ):
        return True

    return False



"""
Types for the licenses used on Flickr.
"""

from datetime import datetime
import typing


class License(typing.TypedDict):
    """
    The license of a particular photo.

    The ID is a human-readable ID chosen by us; the label and URL
    come from Flickr.
    """

    id: "LicenseId"
    label: str
    url: str


LicenseId = typing.Literal[
    "all-rights-reserved",
    "cc-by-nc-sa-2.0",
    "cc-by-nc-2.0",
    "cc-by-nc-nd-2.0",
    "cc-by-2.0",
    "cc-by-sa-2.0",
    "cc-by-nd-2.0",
    "nkcr",
    "usgov",
    "cc0-1.0",
    "pdm",
    "cc-by-4.0",
    "cc-by-sa-4.0",
    "cc-by-nd-4.0",
    "cc-by-nc-4.0",
    "cc-by-nc-sa-4.0",
    "cc-by-nc-nd-4.0",
]


NAME_TO_LICENSE_ID: dict[str, LicenseId] = {
    "All Rights Reserved": "all-rights-reserved",
    "CC BY-NC-SA 2.0": "cc-by-nc-sa-2.0",
    "CC BY-NC 2.0": "cc-by-nc-2.0",
    "CC BY-NC-ND 2.0": "cc-by-nc-nd-2.0",
    "CC BY 2.0": "cc-by-2.0",
    "CC BY-SA 2.0": "cc-by-sa-2.0",
    "CC BY-ND 2.0": "cc-by-nd-2.0",
    "No known copyright restrictions": "nkcr",
    "United States Government Work": "usgov",
    "Public Domain Dedication (CC0)": "cc0-1.0",
    "Public Domain Mark": "pdm",
    "CC BY 4.0": "cc-by-4.0",
    "CC BY-SA 4.0": "cc-by-sa-4.0",
    "CC BY-ND 4.0": "cc-by-nd-4.0",
    "CC BY-NC 4.0": "cc-by-nc-4.0",
    "CC BY-NC-SA 4.0": "cc-by-nc-sa-4.0",
    "CC BY-NC-ND 4.0": "cc-by-nc-nd-4.0",
}


NAME_OVERRIDES = {
    "Public Domain Dedication (CC0)": "CC0 1.0",
}


class LicenseChangeEntry:
    """
    Events in the license history of a photo -- both the initial license
    and any subsequent changes.
    """

    # The initial license, set when the photo was uploaded
    InitialLicense = typing.TypedDict(
        "InitialLicense", {"date_posted": datetime, "license": License}
    )

    # Any changes to the license made after the initial upload
    ChangedLicense = typing.TypedDict(
        "ChangedLicense",
        {"date_changed": datetime, "old_license": License, "new_license": License},
    )


LicenseChange = LicenseChangeEntry.InitialLicense | LicenseChangeEntry.ChangedLicense



from .base import HttpxImplementation as HttpxImplementation
from .comment_methods import CommentMethods
from .commons_methods import FlickrCommonsMethods
from .license_methods import LicenseMethods
from .single_photo_methods import SinglePhotoMethods
from .user_methods import UserMethods


class FlickrApiMethods(
    CommentMethods,
    FlickrCommonsMethods,
    SinglePhotoMethods,
    LicenseMethods,
    UserMethods,
):
    pass


class FlickrApi(HttpxImplementation, FlickrApiMethods):
    pass



"""
Functions for actually downloading a photo -- you get the URL from
the Flickr API, then download the image itself here.
"""

from pathlib import Path
import typing
import uuid

import httpx
from tenacity import retry, stop_after_attempt, wait_exponential, retry_if_exception

from .retrying import is_retryable


default_client = httpx.Client(headers={"User-Agent": "flickr-photos-api"})


class DownloadedFile(typing.TypedDict):
    """
    Represents a downloaded file.
    """

    path: Path
    content_type: str


@retry(
    stop=stop_after_attempt(5),
    wait=wait_exponential(multiplier=1, min=4, max=60),
    retry=retry_if_exception(is_retryable),
)
def download_file(
    client: httpx.Client = default_client,
    *,
    url: str,
    download_dir: Path,
    base_name: str,
) -> DownloadedFile:
    """
    Download a file from Flickr.com.

    This function will retry the download a certain number of times
    if it fails, e.g. if Flickr returns an error or if the download
    is rate-limited.
    """
    download_dir.mkdir(exist_ok=True, parents=True)

    download_path = download_dir / base_name

    # We use a streaming response from HTTPX because some Flickr files
    # can be very big, e.g. original video files.  We don't need or
    # want to buffer the whole thing into memory.
    tmp_path = download_path.with_suffix(f".{uuid.uuid4()}.tmp")

    with client.stream("GET", url, follow_redirects=True) as resp:
        resp.raise_for_status()

        with open(tmp_path, "xb") as out_file:
            for data in resp.iter_bytes():
                out_file.write(data)

    # Now work out the correct file extension to use for this content type.
    #
    # If we can't work out a good file extension, leave it blank.
    content_type = resp.headers["content-type"]

    try:
        file_suffix = {
            "image/jpeg": ".jpg",
            "image/png": ".png",
            "image/gif": ".gif",
            "video/mp4": ".mp4",
        }[content_type]

        out_path = download_path.with_suffix(file_suffix)
    except KeyError:
        out_path = download_path

    # Rename the temporarily downloaded file to the final file.
    tmp_path.rename(out_path)

    return {"path": out_path, "content_type": content_type}



"""
Tests for `flickr_api.models`.
"""

from nitrate.types import validate_type

from flickr_api import FlickrApi
from flickr_api.models import SinglePhoto


def test_validate_single_photo_info(flickr_api: FlickrApi) -> None:
    """
    We can use `SinglePhoto` with `validate_type`.

    We use this helper extensively in our other projects, and it's
    important that it's compatible with this model.
    """
    photo = flickr_api.get_single_photo(photo_id="54159643533")

    assert validate_type(photo, model=SinglePhoto)
